import json
import logging

import click
from rasterio.rio.options import creation_options

from rio_toa.radiance import calculate_landsat_radiance
from rio_toa.reflectance import calculate_landsat_reflectance
from rio_toa.brightness_temp import calculate_landsat_brightness_temperature
from rio_toa.toa_utils import _parse_bands_from_filename, _parse_mtl_txt

logger = logging.getLogger('rio_toa')


@click.group('toa')
def toa():
    """Top of Atmosphere (TOA) correction for landsat 8
    """
    pass


@click.command('radiance')
@click.argument('src_path', type=click.Path(exists=True))
@click.argument('src_mtl', type=click.Path(exists=True))
@click.argument('dst_path', type=click.Path(exists=False))
@click.option('--dst-dtype',
              type=click.Choice(['uint16', 'uint8']),
              default='uint16',
              help='Output data type')
@click.option('--rescale-factor', '-r', type=float,
              default=None,
              help="Rescale TOA values by a multiplier. (Default: "
                   "55000 for uint16, 215 for uint8, 1.0 for float32)")
@click.option('--clip/--no-clip', default=True,
              help="Clip raw TOA values to constrain the domain to 0..1 "
              "(Default: True)")
@click.option('--readtemplate', '-t', default=".*/LC8.*\_B{b}.TIF",
              help="File path template [Default ='.*/LC8.*\_B{b}.TIF']")
@click.option('--workers', '-j', type=int, default=4)
@click.option('--l8-bidx', default=0, type=int,
              help="L8 Band that the src_path represents"
              "(Default is parsed from file name)")
@click.option('--verbose', '-v', is_flag=True, default=False)
@click.pass_context
@creation_options
def radiance(ctx, src_path, src_mtl, dst_path, rescale_factor,
             readtemplate, verbose, creation_options, l8_bidx,
             dst_dtype, workers, clip):
    """Calculates Landsat8 Top of Atmosphere Radiance
    """
    if verbose:
        logger.setLevel(logging.DEBUG)

    if l8_bidx == 0:
        l8_bidx = _parse_bands_from_filename([src_path], readtemplate)[0]

    calculate_landsat_radiance(src_path, src_mtl, dst_path,
                               rescale_factor, creation_options, l8_bidx,
                               dst_dtype, workers, clip)


@click.command('reflectance')
@click.argument('src_paths', nargs=-1, type=click.Path(exists=True))
@click.argument('src_mtl', type=click.Path(exists=True))
@click.argument('dst_path', type=click.Path(exists=False))
@click.option('--dst-dtype',
              type=click.Choice(['uint16', 'uint8', 'float32']),
              default='uint16',
              help='Output data type')
@click.option('--rescale-factor', '-r', type=float,
              default=None,
              help="Rescale TOA values by a multiplier. (Default: "
                   "55000 for uint16, 215 for uint8, 1.0 for float32)")
@click.option('--clip/--no-clip', default=True,
              help="Clip raw TOA values to constrain the domain to 0..1 "
              "(Default: True)")
@click.option('--readtemplate', '-t', default=".*/LC8.*\_B{b}.TIF",
              help="File path template [Default ='.*/LC8.*\_B{b}.TIF']")
@click.option('--workers', '-j', type=int, default=4)
@click.option('--l8-bidx', default=0, type=int,
              help="L8 Band that the src_path represents"
              "(Default is parsed from file name)")
@click.option('--verbose', '-v', is_flag=True, default=False)
@click.option('--pixel-sunangle', '-p', is_flag=True, default=False,
              help="Per pixel sun elevation")
@click.pass_context
@creation_options
def reflectance(ctx, src_paths, src_mtl, dst_path, dst_dtype,
                rescale_factor, clip, readtemplate, workers, l8_bidx,
                verbose, creation_options, pixel_sunangle):
    """Calculates Landsat8 Top of Atmosphere Reflectance
    """
    if verbose:
        logger.setLevel(logging.DEBUG)

    if l8_bidx == 0:
        l8_bidx = _parse_bands_from_filename(list(src_paths), readtemplate)

    calculate_landsat_reflectance(list(src_paths), src_mtl, dst_path,
                                  rescale_factor, creation_options,
                                  list(l8_bidx), dst_dtype,
                                  workers, pixel_sunangle, clip)


@click.command('brighttemp')
@click.argument('src_path', type=click.Path(exists=True))
@click.argument('src_mtl', type=click.Path(exists=True))
@click.argument('dst_path', type=click.Path(exists=False))
@click.option('--dst-dtype', '-d',
              type=click.Choice(['float32', 'float64', 'uint16', 'uint8']),
              default='float32',
              help='Output data type')
@click.option('--temp-scale', '-s',
              type=click.Choice(['K', 'F', 'C']),
              default='K',
              help='Temperature scale [Default = K (Kelvin)]')
@click.option('--readtemplate', '-t', default=".*/LC8.*\_B{b}.TIF",
              help="File path template [Default ='.*/LC8.*\_B{b}.TIF']")
@click.option('--workers', '-j', type=int, default=4)
@click.option('--thermal-bidx', default=0, type=int,
              help="L8 thermal band that the src_path represents"
              "(Default is parsed from file name)")
@click.option('--verbose', '-v', is_flag=True, default=False)
@click.pass_context
@creation_options
def brighttemp(ctx, src_path, src_mtl, dst_path, dst_dtype,
               temp_scale, readtemplate, workers,
               thermal_bidx, verbose, creation_options):
    """Calculates Landsat8 at-satellite brightness temperature.
    TIRS band data can be converted from spectral radiance
    to brightness temperature using the thermal
    constants provided in the metadata file:
    """

    if verbose:
        logger.setLevel(logging.DEBUG)

    if thermal_bidx == 0:
        thermal_bidx = _parse_bands_from_filename([src_path], readtemplate)[0]

    calculate_landsat_brightness_temperature(
        src_path, src_mtl, dst_path, temp_scale,
        creation_options, thermal_bidx, dst_dtype, workers)


@click.command('parsemtl')
@click.argument('mtl', default='-', required=False)
def parsemtl(mtl):
    """Converts a Landsat 8 text MTL
    to JSON
    """
    try:
        mtl = str(click.open_file(mtl).read())
    except IOError:
        mtl = str('\n'.join([inputtiles]))

    click.echo(json.dumps(_parse_mtl_txt(mtl)))


toa.add_command(radiance)
toa.add_command(reflectance)
toa.add_command(brighttemp)
toa.add_command(parsemtl)



import json
import numpy as np
import rasterio as rio
import collections
from rasterio.coords import BoundingBox
import riomucho
from rasterio import warp

from rio_toa import radiance
from rio_toa import toa_utils
from rio_toa import sun_utils


def brightness_temp(img, ML, AL, K1, K2, src_nodata=0):
    """Calculate brightness temperature of Landsat 8
    as outlined here: http://landsat.usgs.gov/Landsat8_Using_Product.php

    T = K2 / np.log((K1 / L)  + 1)

    and

    L = ML * Q + AL

    where:
        T  = At-satellite brightness temperature (degrees kelvin)
        L  = TOA spectral radiance (Watts / (m2 * srad * mm))
        ML = Band-specific multiplicative rescaling factor from the metadata
             (RADIANCE_MULT_BAND_x, where x is the band number)
        AL = Band-specific additive rescaling factor from the metadata
             (RADIANCE_ADD_BAND_x, where x is the band number)
        Q  = Quantized and calibrated standard product pixel values (DN)
             (ndarray img)
        K1 = Band-specific thermal conversion constant from the metadata
             (K1_CONSTANT_BAND_x, where x is the thermal band number)
        K2 = Band-specific thermal conversion constant from the metadata
             (K1_CONSTANT_BAND_x, where x is the thermal band number)


    Parameters
    -----------
    img: ndarray
        array of input pixels
    ML: float
        multiplicative rescaling factor from scene metadata
    AL: float
        additive rescaling factor from scene metadata
    K1: float
        thermal conversion constant from scene metadata
    K2: float
        thermal conversion constant from scene metadata

    Returns
    --------
    ndarray:
        float32 ndarray with shape == input shape
    """
    L = radiance.radiance(img, ML, AL, src_nodata=0)
    L[img == src_nodata] = np.NaN

    T = K2 / np.log((K1 / L) + 1)

    return T


def _brightness_temp_worker(data, window, ij, g_args):
    """rio mucho worker for brightness temperature. It reads input
    files and perform reflectance calculations on each window.

    Parameters
    ------------
    open_files: list of rasterio open files
    window: tuples
    g_args: dictionary

    Returns
    ---------
    out: None
        Output is written to dst_path
    """

    output = toa_utils.temp_rescale(
                    brightness_temp(
                        data[0],
                        g_args['M'],
                        g_args['A'],
                        g_args['K1'],
                        g_args['K2'],
                        g_args['src_nodata']),
                    g_args['temp_scale'])

    return output.astype(g_args['dst_dtype'])


def calculate_landsat_brightness_temperature(
        src_path, src_mtl, dst_path, temp_scale,
        creation_options, band, dst_dtype, processes):

    """Parameters
    ------------
    src_path: list
              list of src_paths(strings)
    src_mtl: string
             mtl file path
    dst_path: string
              destination file path
    rescale_factor: float [default] float(55000.0/2**16)
                    rescale post-TOA tifs to 55,000 or to full 16-bit
    creation_options: dictionary
                      rio.options.creation_options
    band: list
          list of integers
    dst_dtype: strings [default] uint16
               destination data dtype

    Returns
    ---------
    out: None
        Output is written to dst_path
    """
    mtl = toa_utils._load_mtl(src_mtl)

    M = toa_utils._load_mtl_key(mtl,
                                ['L1_METADATA_FILE',
                                 'RADIOMETRIC_RESCALING',
                                 'RADIANCE_MULT_BAND_'],
                                band)
    A = toa_utils._load_mtl_key(mtl,
                                ['L1_METADATA_FILE',
                                 'RADIOMETRIC_RESCALING',
                                 'RADIANCE_ADD_BAND_'],
                                band)

    K1 = toa_utils._load_mtl_key(mtl,
                                 ['L1_METADATA_FILE',
                                  'TIRS_THERMAL_CONSTANTS',
                                  'K1_CONSTANT_BAND_'],
                                 band)
    K2 = toa_utils._load_mtl_key(mtl,
                                 ['L1_METADATA_FILE',
                                  'TIRS_THERMAL_CONSTANTS',
                                  'K2_CONSTANT_BAND_'],
                                 band)

    dst_dtype = np.__dict__[dst_dtype]

    with rio.open(src_path) as src:
        dst_profile = src.profile.copy()

        src_nodata = src.nodata

        for co in creation_options:
            dst_profile[co] = creation_options[co]

        dst_profile['dtype'] = dst_dtype

    global_args = {
        'M': M,
        'A': A,
        'K1': K1,
        'K2': K2,
        'src_nodata': 0,
        'temp_scale': temp_scale,
        'dst_dtype': dst_dtype
        }

    with riomucho.RioMucho([src_path],
                           dst_path,
                           _brightness_temp_worker,
                           options=dst_profile,
                           global_args=global_args) as rm:

        rm.run(processes)
